/*
		Convolution or Recursion Equation
		
    Copyright (C) 2015  Ralph Schenn

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <alloc.h>
#include <args.h>
#include <asmtypes.h>
#include <error.h>
#include <estack.h>
#include <stdlib.h>
#include "Stack.h"

/* Multiply and Accumulate
input or output
length of input or output
coef a or coef b
length of coef a or coef b
index in output */
float MulAcc(float xy[],
	unsigned short lxy,
	float ab[],
	unsigned short lab,
	unsigned short iy)
{
	// input or output index
	short k;
	
	// impulse response or recursion coefficients index
	unsigned short j;
	
	// sum of products
	float s = 0;
	
	// loop over impulse response or recursion coefficients
	for (j = 0; j < lab; j++) {
		k = iy - j;
			
		// valid input or output signal index
		if (0 <= k && k < (short)lxy) {
			s += ab[j] * xy[k]; 
		}
	}
	
	return s;	
}

/* Convolution or Recursive Equation
calculate output signal from input signal using recursion coefficients
input signal
length of input signal 
a coefficients (weight of input signal)
length of a coefficients 
b coefficients (weight of output signal)
length of b coefficients 
output signal */
void ConRecEq(float x[],
	unsigned short xl,
	float a[],
	unsigned short al,
	float b[],
	unsigned short bl,
	float y[])
{  	
	// length of output signal
 	unsigned short yl = xl + al - 1;
 	
 	unsigned short i;
 	 	
	// loop over output array
	for (i = 0; i < yl; i++) {
		// convolve input with impulse response
		y[i] = MulAcc (x,				// input
				xl, 					// length of input
				a, 						// impulse response
				al, 					// length of impulse response
				i) + 					// output index
			// convolve output with recursion coefficients
			(b == NULL ? 0 : MulAcc (y, // output
				yl, 					// length of output
				b, 						// recursion coefs
				bl, 					// length of recursion coefs
				i - 1));				// output index
	}
}

/* Convolution or Recursive Equation for TI-89
pointer to lists of real numbers on estack 
two list=convoltion 
three lists=recursion equation */
void ConRecEqWrapper(CESI argptr)
{
	HANDLE xHdl = H_NULL;
	HANDLE aHdl = H_NULL;
	HANDLE bHdl = H_NULL;
	HANDLE yHdl = H_NULL;

	float *bPtr = NULL;
	
	// false=convolution, true=recursion
	enum Bool recEq = ArgCount () == 3;
	
	TRY	
		// get input signal
		unsigned short lenX = GetFloatListArg (&argptr, &xHdl);
		
		// get impulse response
		unsigned short lenA = GetFloatListArg (&argptr, &aHdl);
		
		unsigned short lenB = 0;
		
		// recursion
		if (recEq) {
			// get recursion coefficients
			lenB = GetFloatListArg (&argptr, &bHdl);
			
			// invalid recursion coefficients
			if (bHdl == H_NULL || lenB == 0) {
				ER_throw (ER_NON_CONFORMING_LISTS); 
			}
			
			bPtr = HeapDeref (bHdl);				
		}
		
		// length of output signal
		unsigned short lenY = lenX + lenA - 1;
				
		// input signal and impulse response is valid
		if (xHdl != H_NULL && // ???
			aHdl != H_NULL && 
			lenX > 0 && 
			lenA > 0) {		
			CleanStack ();
		
			float *xPtr = NULL;
			float *aPtr = NULL;
			float *yPtr = NULL;
		
			xPtr = HeapDeref (xHdl);
			aPtr = HeapDeref (aHdl);
					
			// allocate array for output signal
			yHdl = HeapAlloc (lenY * sizeof (float)); 
		
			// allocation of output signal succeded
			if (yHdl != H_NULL) {
				yPtr = HeapDeref (yHdl);
			
				// bPtr == NULL convolution 
				// bPtr != NULL recursion
				ConRecEq(xPtr,	// input
					lenX, 		// length of input
					aPtr, 		// impulse response
					lenA, 		// length of impulse response
					bPtr, 		// recursion coefs
					lenB, 		// length of recursion coefs
					yPtr); 		// length of output
			
				// push list in reverse order to estack
				push_float_list (yPtr, lenY); 
			}			
		} else {
			ER_throw (ER_DOMAIN);
		}
	FINALLY
		// avoid address error when arga are empty lists
		if (xHdl != H_NULL)
		{
			HeapFree (xHdl);
		}

		if (aHdl != H_NULL)
		{
			HeapFree (aHdl);
		}
		
		if (bHdl != H_NULL)
		{
			HeapFree (bHdl);
		}
		
		if (yHdl != H_NULL)
		{
			HeapFree (yHdl);
		}
	ENDFINAL
}
