//   fft.h - declaration of class
//   of fast Fourier transform - FFT
//
//   The code is property of LIBROW
//   You can use it on your own
//   When utilizing credit LIBROW site
//   
//   The following modifications (c) 2015 Ralph Schenn
//   - port from C++ to C
//   - port to TI-89 pocket calculator
 
#include "FFT.h"
#include <math.h>
#include <alloc.h>
#include <error.h>
#include <estack.h>
#include "Stack.h" 
 
/* Inplace version of rearrange function */
void Rearrange(_Complex float* Data, 
	const unsigned short N)
{
	//   Swap position
	unsigned short Target = 0;

 	unsigned short Position;
 				
	//   Process all positions of input signal
	for (Position = 0; Position < N; ++Position)
	{
		//   Only for not yet swapped entries
		if (Target > Position)
		{
			//   Swap entries
			const _Complex float Temp = Data[Target];
			Data[Target] = Data[Position];
			Data[Position] = Temp;
		}
 
		//   Bit mask
		unsigned short Mask = N;
 
		//   While bit is set
        while (Target & (Mask >>= 1))
		{
			//   Drop bit
			Target &= ~Mask;
		}
 
		//   The current bit is 0 - set it
		Target |= Mask;
	}
}
 
/* FFT implementation */
void Perform(_Complex float* Data, 
	const unsigned short N, 
	const enum Bool Inverse)
{
	const float pi = Inverse ? PI : -PI;
 
	unsigned short Step;
 				
	//   Iteration through dyads, quadruples, octads and so on...
	for (Step = 1; Step < N; Step <<= 1)
	{
		//   Jump to the next entry of the same transform factor
		const unsigned short Jump = Step << 1;
 
		//   Angle increment
		const float delta = pi / (float)Step;
 
		//   Auxiliary sin(delta / 2)
		const float Sine = sin(delta * .5);
 
		//   Multiplier for trigonometric recurrence
		const _Complex float Multiplier = -2. * Sine * Sine + sin(delta) * 1.0fi;
 
		//   Start value for transform factor, fi = 0
		_Complex float Factor = 1.;
		
		unsigned short Group;
 								
		//   Iteration through groups of different transform factor
		for (Group = 0; Group < Step; ++Group)
		{
			unsigned short Pair;
                        
			//   Iteration within group
			for (Pair = Group; Pair < N; Pair += Jump)
			{
				//   Match position
				const unsigned short Match = Pair + Step;
 
				//   Second term of two-point transform
				const _Complex float Product = Factor * Data[Match];
 
				//   Transform for fi + pi
				Data[Match] = Data[Pair] - Product;
 
				//   Transform for fi
				Data[Pair] += Product;
			}
 
			//   Successive transform factor via trigonometric recurrence
			Factor = Multiplier * Factor + Factor;
		}
	}
}
 
/* Scaling of inverse FFT result */
void ScaleData(_Complex float* Data, 
	const unsigned short N)
{
	const float Factor = 1. / (float)N;
        
	unsigned short Position;
 
	//   Scale all data entries
	for (Position = 0; Position < N; ++Position)
	{
		Data[Position] *= Factor;
	}
}

/* FOURIER TRANSFORM, INPLACE VERSION
Data  - both input data and output
N     - length of both input data and result */
enum Bool FFT(_Complex float* Data,
	const unsigned short N,
	const enum Bool Inverse)
{
	//   Check input parameters
	if (!Data || N < 1 || N & (N - 1))
	{
		return FALSE;
	}
 
	//   Rearrange
	Rearrange(Data, N);
 
	//   Call FFT implementation
	Perform(Data, N, Inverse);
 
	//   Scale if necessary
	if (Inverse)
	{
		ScaleData(Data, N);
	}
 
	//   Succeeded
	return TRUE;
}

/* FOURIER TRANSFORM FOR TI-89
argptr  - pointer to list of complex numbers on estack
inverse - inverse transform */
void FFTWrapper(CESI argptr, 
	enum Bool inverse)
{
	HANDLE handle = H_NULL;

	TRY	
		// read list into array
		unsigned short l = GetComplexFloatListArg (&argptr, &handle);
		
		CleanStack ();
		
		_Complex float *array = NULL;
		
		// list is not empty
		if (handle != H_NULL)
		{
			array = HeapDeref (handle);
		
	  		if (FFT (array, l, inverse))
	  		{
				// push list in reverse order to estack
				push_complex_float_list (array, l); 			
	  		}
	  		else
	  		{
	  			ER_throw (ER_DIMENSION);
	  		}
		// passed list is empty
	  	} else {
	  		ER_throw (NON_CONFORMING_LISTS_ERROR);	
	  	}
	FINALLY
		// avoid address error when arg is empty list
		if (handle != H_NULL)
		{
			HeapFree (handle);
		}
	ENDFINAL
}
