/*
		Utilities for Estack Interaction
		
    Copyright (C) 2015  Ralph Schenn with support from Kevin Kofler

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#define MIN_AMS 209

/* convert bool value to corresponding tag value */
#define bool_to_tag(b) ((b) + 43)

#include <alloc.h>
#include <args.h>
#include <error.h>
#include <estack.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "Stack.h"
#include "Tags.h"

/* check if char argument is contained in list
returns the value of the argument at position argPos
??? this function should also be implemented for strings */
const char *CheckArgCharVal(unsigned short argPos, 
	char *allowedVal)
{
	// argument at position does not exist
	if (argPos > ArgCount ()) {
		ER_throw (ER_TOO_FEW_ARGS);		
	}
	
	// get pointer to argument at position
	CESI argptr = EX_getArg (argPos);
	
	// argument has wrong type
	if (GetArgType (argptr) != STR_TAG) {
		ER_throw (ER_ARG_MUST_BE_STRING);		
	}
			
	// get value of argument at position
	const char *arg = GetStrnArg (argptr);
				
	// check if val is not empty string and contained in string
	if (strstr (allowedVal, arg) == NULL || !strlen (arg) ) {
		ER_throw (ER_DOMAIN);
	}
	
	// return value of argument at position
	return arg;
}

/* check if number of arguments is correct */
char CheckNumArgs(unsigned char expArgs)
{
	unsigned char args = ArgCount ();
	
	// check number of arguments
	if (args < expArgs) {
		ER_throw (ER_TOO_FEW_ARGS);
	} else if (args > expArgs) {
		ER_throw (ER_TOO_MANY_ARGS);
	}	
	
	return args;
}

/* check if number of arguments is correct */
char CheckNumArgsInt(unsigned char minExpArgs, 
	unsigned char maxExpArgs)
{
	unsigned char args = ArgCount ();
	
	// check number of arguments
	if (args < minExpArgs) {
		ER_throw (ER_TOO_FEW_ARGS);
	} else if (args > maxExpArgs) {
		ER_throw (ER_TOO_MANY_ARGS);
	}	
	
	return args;
}

/* remove arguments passed to program from stack */
void CleanStack() // ok
{
	// iterate from top to bottom of stack
	// (first argument to last argument)
	while (GetArgType (top_estack) != END_TAG) {
		// get next argument
  		top_estack = next_expression_index (top_estack);		
	}
	
	top_estack--;
}

/* return type of list 
only the first cell is checked */
enum Tags ListType(CESI ptr)
{
	if (GetArgType (ptr) != LIST_TAG) {
		ER_throw (ER_ARGUMENT_MUST_BE_LIST);
	}

	return GetArgType (ptr - 1);	
}

/* return type of matrix 
only the first cell is checked */
enum Tags MatrixType(CESI ptr)
{
	if (!IsMatrix (ptr)) {
		ER_throw (ER_ARG_MUST_BE_MATRIX);
	}
	
	return GetArgType (ptr - 2);
}

/* reverse the order of the array elements 
??? additional arg for length */
void ReverseArray(CESI array[])
{
	// length of array
	unsigned short l = sizeof(array) / sizeof(CESI);
	
	// counting upward
	unsigned short i;
	
	// counting downward
	unsigned short j = l - 1;
	
	// temp storage for swapping
	CESI t;
	
	// loop over all elements
	for (i = 0; i < l; i++) {
		t = array[i]; // ??? use macro for swapping
		array[i] = array[j];
		array[j] = t;
		j--;
	}	
}

/* determine begin of stack */
void StackBegin(ESI *ptr) // 
{
	// iterate from top to bottom of stack
	// (first argument to last argument)
	while (GetArgType (*ptr) != END_TAG) {
		// get next argument
  		*ptr = next_expression_index (*ptr);		
	}
	
  (*ptr)--;
}

// G E T   F U N C T I O N S //////////////////////////////////////////////////

/* get int, float or complex from estack */
_Complex float estack_number_to_Complex(CESI ptr)
{
	switch (GetArgType (ptr))
	{
		// argument is scalar
		case FLOAT_TAG:
		case NEGINT_TAG:
		case POSINT_TAG:
		case NEGINFINITY_TAG:
		case INFINITY_TAG:
		case UNDEF_TAG:
			return estack_number_to_Float_inf (ptr);
		// argument is complex	
		case COMPLEX_TAG:
			return GetComplexFloatArg (&ptr);
		default:
			ER_throw (ER_DATATYPE);
	}
}

/* get type of argument with number (0-based index) */
enum Tags GetArgNumType(short argNum)
{
	return GetArgType (EX_getArg (argNum));
}

/* get complex float from estack (16.7)
??? handle polar format */
_Complex float GetComplexFloatArg(CESI *ptr) // 
{
	_Complex float input;

	//  ... FLOAT_TAG ... FLOAT_TAG COMPLEX_TAG
	
	// check if expression is complex
	// (this is internal representation)
	if (GetArgType (*ptr) != COMPLEX_TAG) {
		ER_throw (ER_DATATYPE);
	}	
	
	// skip the COMPLEX_TAG
	(*ptr)--;
	
	// get imaginary part
	// without - the sign gets inverted	
	__imag__ input = estack_number_to_Float_inf (*ptr);
	*ptr = next_expression_index (*ptr);
		
	// get real part
	__real__ input = estack_number_to_Float_inf (*ptr);
	*ptr = next_expression_index (*ptr);
		
	return input;
}

/* get float from expression stack */
float estack_number_to_Float_inf(CESI ptr) // ok
{
	switch (GetArgType (ptr)) {
		case NEGINFINITY_TAG:
			return NEGATIVE_INF;
		case INFINITY_TAG:
			return POSITIVE_INF;		
		case UNDEF_TAG:
			return NAN;
		case FLOAT_TAG:
		case POSINT_TAG:
		case NEGINT_TAG:
			// estack_number_to_Float does not handle NAN
			return estack_number_to_Float (ptr);
		default:
			ER_throw (ER_DATATYPE); 
	}
}

/* get list of floats 
content is written to output (size of array is adjusted automatically)
returns number of elements */
unsigned short GetComplexFloatListArg(CESI *ptr, 
	HANDLE *output) // ok
{
	// number of elements in list
	unsigned short i = 0;
	
	// wrong argument type
	if (GetArgType (*ptr) != LIST_TAG) {
		ER_throw (ER_ARGUMENT_MUST_BE_LIST);		
	}
	
	// move index to first element of list
	// (without decrement we start at END_TAG and are not looping)
	(*ptr)--;
	
	// while not at end of list
	while (GetArgType (*ptr) != END_TAG) {
		// allocate space for additional list element
		*output = HeapReallocThrow(*output, (i + 1) * sizeof(_Complex float));

		_Complex float *array = HeapDeref (*output);

		// convert to float and copy to array
		// (this also advances the pointer to the next list element -
		// do not call next_expression_index)			
		array[i++] = estack_number_to_Complex (*ptr);
		
		*ptr = next_expression_index (*ptr);
	}
	
	// point to next argument instead of LIST_END_TAG
	// (otherwise getting next argument from estack will fail)
	(*ptr)--;
	
	return i;	
}
 
/* get list of floats 
content is written to output (size of array is adjusted automatically)
returns number of elements */
unsigned short GetFloatListArg(CESI *ptr, 
	HANDLE *output) // ok
{	
	// number of elements in list
	unsigned short i = 0;
	
	// wrong argument type
	if (GetArgType (*ptr) != LIST_TAG) {
		ER_throw (ER_ARGUMENT_MUST_BE_LIST);		
	}
	
	// move index to first element of list
	// (without decrement we start at END_TAG and are not looping)
	(*ptr)--;

	// while not at end of list
	while (GetArgType (*ptr) != END_TAG) {
		// allocate space for additional list element
		*output = HeapReallocThrow(*output, (i + 1) * sizeof(float));
	
		float *array = HeapDeref (*output);
		
		// convert to float and copy to array
		// (in contrast with GetIntArg and GetFloatArg this
		// function does not change the pointer to point
		// to the next argument)
		array[i++] = estack_number_to_Float_inf (*ptr);
		
		// advance pointer to next element
		*ptr = next_expression_index (*ptr);			
	}
	
	// point to next argument instead of LIST_END_TAG
	// (otherwise getting next argument from estack will fail)
	(*ptr)--;

	return i;
}

/* count the number of elements in list */
unsigned short GetListDim(CESI ptr)
{	
	// number of list elements
	unsigned short i = 0;
	
	if (GetArgType (ptr) != LIST_TAG) {
		ER_throw (ER_ARGUMENT_MUST_BE_LIST);
	}
	
	// move index to first element of list
	// (without decrement we start at END_TAG and are not looping)
	--ptr;

	// while not at end of list
	while (GetArgType (ptr) != END_TAG) { 
		// count the element
		i++;
		
		// step to next list element
		ptr = next_expression_index (ptr);
	}

	return i;
}

/* get list of longints 
content is written to output (size of array is adjusted automatically)
returns number of elements */
unsigned short GetLongintListArg(CESI *ptr, 
	HANDLE *output)
{	
	// number of elements in list
	unsigned short i = 0;
	unsigned char type;
	
	// wrong argument type
	if (GetArgType (*ptr) != LIST_TAG) {
		ER_throw (ER_ARGUMENT_MUST_BE_LIST);		
	}
	
	// move index to first element of list
	// (without decrement we start at END_TAG and are not looping)
	(*ptr)--;

	// while not at end of list
	while ((type = GetArgType (*ptr)) != END_TAG) {
		// list element is float
		if (type == NEGINT_TAG || 
			type == POSINT_TAG) {			
			// allocate space for additional list element
			*output = HeapReallocThrow(*output, (i + 1) * sizeof(long));
		
			// convert to longint and copy to array
			// (this also advances the pointer to the next list element -
			// do not call next_expression_index)
			long *array = HeapDeref (*output);
			
			array[i++] = GetSignIntArg (ptr);
		// list element is not float
		} else {			
			ER_throw (ER_DATATYPE);			
		}
	}

	// point to next argument instead of LIST_END_TAG
	(*ptr)--;

	return i;
}

/* get matrix of floats 
content is written to output (size of array is adjusted automatically)
also returns number of rows and colums
{ { ELEM... } { ELEM... } }
LIST_TAG LIST_TAG ELEM... END_TAG LIST_TAG ELEM... END_TAG END_TAG */
void GetFloatMatArg(CESI *argptr, // ok
	HANDLE *output, 
	unsigned short *rows,
	unsigned short *cols)
{
	HANDLE rowHdl = H_NULL;
	float *rowPtr;
	float *matPtr;
	unsigned short rowsize; 
	unsigned short type; 
	
	// argument is not matrix
	if (!is_matrix (*argptr)) {
		ER_throw (ER_ARG_MUST_BE_MATRIX);				
	}
		
	// move index to first list=row of matrix
	(*argptr)--;
	
	*rows = 0;
	*cols = 0;
	
	// while not at end of matrix
	while ((type = GetArgType (*argptr)) != END_TAG &&
		type != END_TAG) {
		// get next row of matrix
		*cols = GetFloatListArg (argptr, &rowHdl);
		
		// memory size required to hold one row
		rowsize = (*cols) * sizeof (float);
		
		// allocate space for the next row
		*output = HeapReallocThrow(*output, ((*rows) + 1) * rowsize);
		
		// get pointers from handles
		rowPtr = HeapDeref (rowHdl);
		matPtr = HeapDeref (*output);
		
		// copy actual row to matrix with previous rows
		memcpy (matPtr + (*rows) * (*cols),	// destination
			rowPtr, 						// source
			rowsize);						// length	
			
		(*rows)++;	
	}
	
	// point to next argument instead of LIST_END_TAG
	// (otherwise getting next argument from estack will fail)
	(*argptr)--;
}

/* get matrix of complex floats 
content is written to output (size of array is adjusted automatically)
also returns number of rows and colums
{ { ELEM... } { ELEM... } }
LIST_TAG LIST_TAG ELEM... END_TAG LIST_TAG ELEM... END_TAG END_TAG */
void GetComplexFloatMatArg(CESI *argptr, // ok
	HANDLE *output, 
	unsigned short *rows,
	unsigned short *cols)
{
	HANDLE rowHdl = H_NULL;
	_Complex float *rowPtr;
	_Complex float *matPtr;
	unsigned short rowsize; 
	unsigned short type; 
	
	// argument is not matrix
	if (!is_matrix (*argptr)) {
		ER_throw (ER_ARG_MUST_BE_MATRIX);				
	}
		
	// move index to first list=row of matrix
	(*argptr)--;
	
	*rows = 0;
	*cols = 0;
	
	// while not at end of matrix
	while ((type = GetArgType (*argptr)) != END_TAG &&
		type != END_TAG) {
		// get next row of matrix
		*cols = GetComplexFloatListArg (argptr, &rowHdl);
		
		// memory size required to hold one row
		rowsize = (*cols) * sizeof (_Complex float);
		
		// allocate space for the next row
		*output = HeapReallocThrow(*output, ((*rows) + 1) * rowsize);
		
		// get pointers from handles
		rowPtr = HeapDeref (rowHdl);
		matPtr = HeapDeref (*output);
		
		// copy actual row to matrix with previous rows
		memcpy (matPtr + (*rows) * (*cols),	// destination
			rowPtr, 						// source
			rowsize);						// length	
			
		(*rows)++;	
	}

	// point to next argument instead of LIST_END_TAG
	// (otherwise getting next argument from estack will fail)
	(*argptr)--;
}

/* get signed int from estack */
int GetSignIntArg(CESI *ptr) // ok
{
	switch (GetArgType (*ptr)) { 
		case NEGINT_TAG: 
			return -GetIntArg (*ptr);
			break;
		case POSINT_TAG: 
			return GetIntArg (*ptr);
			break;
		default:
			ER_throw (ER_DATATYPE);
			break;
	}
}

/* get signed long from estack */
int GetSignLongArg(CESI *ptr)  // ok
{
	switch (GetArgType (*ptr)) { 
		case NEGINT_TAG: 
			return -GetLongLongArg (*ptr);
			break;
		case POSINT_TAG: 
			return GetLongLongArg (*ptr);
			break;
		default:
			ER_throw (ER_DATATYPE);
			break;
	}
}

/* get list of strings
content is written to output (size of array is adjusted automatically)
returns number of elements */
unsigned short GetStringListArg(CESI *ptr, 
	HANDLE *output) // ok
{	
	// number of elements in list
	unsigned short i = 0;
	unsigned char type;
	
	// wrong argument type
	if (GetArgType (*ptr) != LIST_TAG) {
		ER_throw (ER_ARGUMENT_MUST_BE_LIST);		
	}
	
	// move index to first element of list
	// (without decrement we start at END_TAG and are not looping)
	(*ptr)--;

	// while not at end of list
	while ((type = GetArgType (*ptr)) != END_TAG) {
		// list element is string
		if (type == STR_TAG) {	
			// allocate space for additional list element
			*output = HeapReallocThrow(*output, (i + 1) * sizeof(char *));

			char **array = HeapDeref (*output);		

			// copy pointer to string to array 
			// (this also advances the pointer to the next list element -
			// do not call next_expression_index
			// content of string is not copied and still only on estack 
			// so list may be removed from stack only after the handle 
			// argument this function changes is no longer needed)
			array[i++] = (char *)GetStrnArg (*ptr);
		// list element is not string
		} else {			
			ER_throw (ER_ARG_MUST_BE_STRING);			
		}
	}

	// point to next argument instead of LIST_END_TAG
	(*ptr)--;

	return i;
}

/* get string argument from estack and copy it to another variable
(to avoid loosing its content while overwriting the stack) */
void GetStrnArgCopy(CESI *ptr, char *dest, 
	unsigned char destLen)
{
	// wrong argument type
	if (GetArgType (*ptr) != STR_TAG) {
		ER_throw (ER_ARG_MUST_BE_STRING);		
	}

	// get search string from estack
	const char *str = GetStrnArg (*ptr);
		
	// string is to long
	if (strlen (str) > destLen) {
		// exit to avoid buffer overflow
		ER_throw (ER_DOMAIN);
	}
		
	// copy the string to a place independent from estack
	// (to avoid overwriting while pushing something on the stack)
	strcpy (dest, str);
}

/* get string argument at position 
??? should also be implemented for other get functions */
const char *GetStrnArgPos(short pos)
{
	CESI ptr = EX_getArg (pos);
	return GetStrnArg (ptr);	
}

/* get signed int argument at position */
int GetSignIntArgPos(short pos)
{
	CESI ptr = EX_getArg (pos);
	return GetSignIntArg (&ptr);	
}

// I S  F U N C T I O N S /////////////////////////////////////////////////////

/* check if ptr points to a matrix */
enum Bool IsMatrix(CESI ptr)
{
	// argument is not matrix
	return GetArgType (ptr) == LIST_TAG && 
	GetArgType (ptr - 1) == LIST_TAG;	
}

/* check if expression is constant */
void push_is_constant(CESI ptr) // ok
{
	// check number of arguments
	CheckNumArgs (1);
		
	// check if argument is constant
	short c = is_constant (ptr);
	
	// remove argument from stack
	CleanStack ();
	
	// push result to estack
	push_quantum (bool_to_tag (c));
}

/* check if expression is free of tag */
void push_is_free_of_tag(CESI ptr)
{
	// check number of arguments
	CheckNumArgs (2);
	
	// check if expression is independant of tag
	short f = is_free_of_tag (ptr, *next_expression_index (ptr));
	
	// remove argument from stack
	CleanStack ();
	
	// push result to estack
	push_quantum (bool_to_tag (f));
}

/* check if expression is independant of variable or expression */
void push_is_independent_of(CESI ptr) // ok
{
	// check number of arguments
	CheckNumArgs (2);
	
	// check if expression is independant of variable or expression
	short i = is_independent_of (ptr, next_expression_index (ptr));
	
	// remove argument from stack
	CleanStack ();
	
	// push result to estack
	push_quantum (bool_to_tag (i));
}

/* check if expression is negative */
void push_is_negative(CESI ptr) // ok
{
	// check number of arguments
	CheckNumArgs (1);
	
	// check if expression is negative
	short i = is_negative (ptr);
	
	// remove argument from stack
	CleanStack ();
	
	// push result to estack
	push_quantum (bool_to_tag (i));
}

/* check if expression is never zero */
void push_is_never0(CESI ptr) // ok
{
	// check number of arguments
	CheckNumArgs (1);
	
	// check if expression is independant of variable
	short i = is_never0 (ptr);
	
	// remove argument from stack
	CleanStack ();
	
	// push result to estack
	push_quantum (bool_to_tag (i));
}

/* check if expression is positive */
void push_is_positive(CESI ptr) // ok
{
	// check number of arguments
	CheckNumArgs (1);
		
	// check if expression is positive
	short i = is_positive (ptr);
	
	// remove argument from stack
	CleanStack ();
	
	// push result to estack
	push_quantum (bool_to_tag (i));
}

/* check if expression is a variable */
void push_is_variable(CESI ptr) // 
{
	// check number of arguments
	CheckNumArgs (1);
		
	// check if expression is a variable
	short i = is_variable (ptr);
	
	// remove argument from stack
	CleanStack ();
	
	// push result to estack
	push_quantum (bool_to_tag (i));
}

// P U S H  F U N C T I O N S  ////////////////////////////////////////////////

/* push string to estack */
void push_AMS_string(CESI ptr)
{
	push_expr_quantum (ptr, STR_TAG);	
}

/* push complex float to estack (16.7) 
??? handle polar format */
void push_complex_float(_Complex float input) // ok
{
	// ... FLOAT_TAG ... FLOAT_TAG IM_TAG MUL_TAG ADD_TAG

	// when imag part is not zero
	if (__imag__ input != 0) {
		push_float_inf (__imag__ input);		
		push_quantum (IM_TAG);
		push_quantum (MUL_TAG);		
	}
	
	if (__real__ input != 0 || 
		(__real__ input == 0 && __imag__ input == 0)) {
		// we are pushing a float
		push_float_inf (__real__ input);
	}
	
	// and not a complex float
	if (__real__ input != 0 && __imag__ input != 0) {
		push_quantum (ADD_TAG);				
	}
}

/* push float allowing infinities 
pushes NAN as undef without special handling */
void push_float_inf(float f) // ok
{
	// <= does not work
	if (f == NEGATIVE_INF) {
		push_quantum (NEGINFINITY_TAG);
	// >= does not work
	} else if (f == POSITIVE_INF) {
		push_quantum (INFINITY_TAG);			
	} else {
		// push_Float handles NAN correctly
		push_Float (f);	
	}
}

/* push list of complex floats to estack */
void push_complex_float_list(_Complex float input[], 
	short length) // ok
{
	// end of list
	push_END_TAG ();
  
	// loop over float array
	while (--length >= 0) {
		push_complex_float (input[length]);
	}
  
	// beginning of list
	push_LIST_TAG ();		
}

/* push matrix of complex floats to estack */
void push_complex_float_matrix(_Complex float input[], 
	short rows, 
	unsigned short cols) // ok
{
	short col;

	// end of matrix
	push_END_TAG ();
  
	// loop over rows
	while (--rows >= 0) {
		// end of row
		push_END_TAG ();
	  
		// loop over columns
  		for (col = cols - 1; col >= 0; col--) {
  			push_complex_float (input[rows * cols + col]);
  		}

		// beginning of row
		push_LIST_TAG ();	
  	}
  
	// beginning of matrix
	push_LIST_TAG ();	
}

/* push empty list */
void push_empty_list(void) // 
{
	// end of list
	push_END_TAG ();
    
	// beginning of list
	push_LIST_TAG ();	
}

/* push array of floats to estack creating list 
pushes NAN as undefined without requiriring special handling */
void push_float_list(float input[], 
	short length) // ok
{
	// end of list
	push_END_TAG ();
  
	// loop over float array
	while (--length >= 0) {
		push_float_inf (input[length]);
	}
  
	// beginning of list
	push_LIST_TAG ();	
}

/* push matrix of floats to estack */
void push_float_mat(float input[][], 
	short rows, 
	unsigned short cols) // ok
{
  	// end of matrix
	push_END_TAG ();
  
	// loop over rows
 	while (--rows >= 0) {	
		push_float_list ((float *)input + rows * cols, cols); 
	}
  
	// beginning of matrix
	push_LIST_TAG ();	
}

/* push matrix of complex floats to estack */
void push_complex_float_mat(_Complex float input[][], 
	short rows, 
	unsigned short cols) // ok
{
  	// end of matrix
	push_END_TAG ();
  
	// loop over rows
 	while (--rows >= 0) {	
		push_complex_float_list ((_Complex float *)input + rows * cols, cols); 
	}
  
	// beginning of matrix
	push_LIST_TAG ();	
}

/* push array of floats to estack creating list in reverse order */
void push_float_reverse_list(float input[], 
	unsigned short length)
{
	unsigned short i;

	// end of list
	push_END_TAG ();
  
	// loop over float array
	for (i = 0; i < length; i++) {
		push_Float (input[i]);
	}
  
	// beginning of list
	push_LIST_TAG ();	
}

/* push array of expressions to estack creating list */
void push_list(CESI input[], 
	short length)
{
	// end of list
	push_END_TAG ();
  
	// loop over float array
	while (--length >= 0) {
		push_expression (input[length]);
	}
  
	// beginning of list
	push_LIST_TAG ();	
}

/* push array of longint to the estack */
void push_longint_list(long input[], 
	short length)
{
	// end of list
	push_END_TAG ();
  
	// loop over float array
	while (--length >= 0) {
		push_longint (input[length]);
	}
  
	// beginning of list
 	push_LIST_TAG ();	
}

/* push array of strings to estack creating list */
void push_string_list(char *input[], 
	short length) // ok
{
	// end of list
	push_END_TAG ();
  
	// loop over string pointer array
	while (--length >= 0) {
		push_ANSI_string (input[length]);
	}
  
	// beginning of list
	push_LIST_TAG ();	
}

/* push array of strings to estack creating list */
void push_string_list_handle(HANDLE input[], 
	short length) // ok
{
	// end of list
	push_END_TAG ();
  
	// loop over string pointer array
	while (--length >= 0) {
		push_ANSI_string ( HeapDeref (input[length]));
	}
  
	// beginning of list
	push_LIST_TAG ();	
}
