/*
		Utilities for Estack Interaction
		
    Copyright (C) 2015  Ralph Schenn with support from Kevin Kofler

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _STACK_H_
#define _STACK_H_

#include <estack.h>

/* check if char argument is contained in list
returns the value of the argument at position argPos */
const char *CheckArgCharVal(unsigned short argPos, 
	char *allowedVal);

/* check if number of arguments is correct */
char CheckNumArgs(unsigned char expArgs);

/* check if number of arguments is correct */
char CheckNumArgsInt(unsigned char minExpArgs, 
	unsigned char maxExpArgs);

/*remove arguments passed to program from stack
(pass top_estack as argument)*/
void CleanStack();

/* return type of list 
only the first cell is checked */
enum Tags ListType(CESI ptr);

/* return type of matrix 
only the first cell is checked */
enum Tags MatrixType(CESI ptr);

/* reverse the order of the array elements */
void ReverseArray(CESI array[]);

/* determine begin of stack */
void StackBegin(ESI *ptr);

// G E T  F U N C T I O N S ///////////////////////////////////////////////////

/* get int, float or complex from estack */
_Complex float estack_number_to_Complex(CESI ptr);

/* get type of argument with number (0-based index) */
enum Tags GetArgNumType(short argNum);

/* get complex float from estack */
_Complex float GetComplexFloatArg(CESI *ptr);

/* get float from expression stack */
float estack_number_to_Float_inf(CESI ptr);

/* get list of floats 
content is written to output (size of array is adjusted automatically)
returns number of elements */
unsigned short GetComplexFloatListArg(CESI *ptr, 
	HANDLE *output);

/* get list of floats 
content is written to output (size of array is adjusted automatically)
returns number of elements */
unsigned short GetFloatListArg(CESI *ptr, 
	HANDLE *output);

/* get list of longints 
content is written to output (size of array is adjusted automatically)
returns number of elements */
unsigned short GetLongintListArg(CESI *ptr, 
	HANDLE *output);

/* get matrix of floats 
content is written to output (size of array is adjusted automatically)
also returns number of rows and colums */
void GetFloatMatArg(CESI *argptr, 
	HANDLE *output, 
	unsigned short *rows,
	unsigned short *cols);

/* get matrix of complex floats 
content is written to output (size of array is adjusted automatically)
also returns number of rows and colums
{ { ELEM... } { ELEM... } }
LIST_TAG LIST_TAG ELEM... END_TAG LIST_TAG ELEM... END_TAG END_TAG */
void GetComplexFloatMatArg(CESI *argptr, // ok
	HANDLE *output, 
	unsigned short *rows,
	unsigned short *cols);

/* count the number of elements in list */
unsigned short GetListDim(CESI ptr);

/* get signed int from estack */
int GetSignIntArg(CESI *ptr);

/* get signed long from estack */
int GetSignLongArg(CESI *ptr);

/* get list of strings
content is written to output (size of array is adjusted automatically)
returns number of elements */
unsigned short GetStringListArg(CESI *ptr, 
	HANDLE *output);

/* get string argument and copy it to another variable
(to avoid loosing its content while overwriting the stack) */
void GetStrnArgCopy(CESI *ptr, char *dest, 
	unsigned char destLen);

/* get string argument at position */
const char *GetStrnArgPos(short pos);

/* get signed int argument at position */
int GetSignIntArgPos(short pos);

// I S  F U N C T I O N S /////////////////////////////////////////////////////

/* check if ptr points to a matrix */
enum Bool IsMatrix(CESI ptr);

// P U S H  F U N C T I O N S /////////////////////////////////////////////////

/* push string to estack */
void push_AMS_string(CESI ptr);

/* push complex float to estack (16.7) */
void push_complex_float(_Complex float input);

/* push float allowing infinities */
void push_float_inf(float f);

/* push matrix of complex floats to estack */
void push_complex_float_list(_Complex float input[], 
	short length);

/* push matrix of complex floats to estack */
void push_complex_float_matrix(_Complex float input[], 
		short rows, 
		unsigned short cols);

/* push empty list */
void push_empty_list(void);

/* push array of floats to estack creating list */
void push_float_list(float input[], 
	short length);

/* push matrix of floats to estack */
void push_float_mat(float input[][], 
	short rows, 
	unsigned short cols);
	
/* push matrix of complex floats to estack */
void push_complex_float_mat(_Complex float input[][], 
	short rows, 
	unsigned short cols);

/* push array of floats to estack creating list in reverse order */
void push_float_reverse_list(float input[], 
	unsigned short length);

/* push array of expressions to estack creating list */
void push_list(CESI input[], 
	short length);

/* push array of longint to the estack */
void push_longint_list(long input[], 
	short length);

/* push array of strings to estack creating list */
void push_string_list(char *input[], 
	short length);

/* push array of strings to estack creating list */
void push_string_list_handle(HANDLE input[], 
	short length);

// W R A P P E R  F U N C T I O N S ///////////////////////////////////////////

/* check if expression is constant and push result on stack */
void push_is_constant(CESI ptr);

/* check if expression is free of tag */
void push_is_free_of_tag(CESI ptr);

/* check if expression is independant of variable or expression */
void push_is_independent_of(CESI ptr);

/* check if expression is never zero */
void push_is_never0(CESI ptr);

/* check if expression is negative */
void push_is_negative(CESI ptr);

/* check if expression is positive */
void push_is_positive(CESI ptr);

/* check if expression is a variable */
void push_is_variable(CESI ptr);

#endif
