/*
		RecStrSe - find variables containing string
		Example: RecStrSe("folder","str")
		if the first parameter is ommited all folders are searched
		
    Copyright (C) 2013  Ralph Schenn with support from Kevin Kofler

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <alloc.h>
#include <error.h>
#include <estack.h>
#include <statline.h>
#include <string.h>
#include <vat.h>
#include "MultiExpr.h"
#include "Search.h"
#include "Stack.h"

#define BUF_LENGTH 30
#define MAX_VAR_LENGTH 8
#define HOME "\x7F"

void _main(void)
{	
	// search folder name
	char folder[MAX_VAR_LENGTH];

	// search string
	char str[BUF_LENGTH];	

	// parent folder of variable
	const char *varFolder;
	
	// flags for variable search
	unsigned short searchFlags;

	// pointer to variable
	SYM_ENTRY *symptr;

	// number of variables which have been searched
	unsigned short numVars = 0;

	// output buffer
	char out[BUF_LENGTH];

	CESI argptr = top_estack;

	// check number of arguments
	char numArgs = CheckNumArgsInt (1, 2);

	TRY
		// search through all folders
		if (numArgs == 1) {
			// search all folders
			searchFlags = FO_RECURSE;
		// search only in  folder given in second argument
		} else {
			// get pointer to folder name from estack
			// copy the string to a variable independent from estack
			// (to avoid overwriting while pushing strings to the list)		
			GetStrnArgCopy (&argptr, folder, MAX_VAR_LENGTH);

			// search only in specified folder
			searchFlags = FO_SINGLE_FOLDER;
		}
		
		// get search string from estack
		// copy the string to a variable independent from estack
		// (to avoid overwriting while pushing strings to the list)		
		GetStrnArgCopy (&argptr, str, BUF_LENGTH);
	ONERR
		ER_throw (ER_DATATYPE);
	ENDTRY
		
	// lock home folder
	FolderOp (SYMSTR (HOME), FOP_LOCK);
	
	symptr = SymFindFirst (SYMSTR (folder), searchFlags);
	
	// remove arguents from estack
	// (symptr points to the estack, so this should be done 
	// only after using this pointer)
	CleanStack ();

	// beginning of list
  push_END_TAG ();

	// at least one variable exists
  if (symptr) {
    // loop over variables in folder(s)
    do {
			// get pointer to variable
			void *var = HeapDeref (symptr->handle);
		
			// determine actual folder
			// (SymFindFolderName does not work when searching only 
			// for the objects in one folder)
			varFolder = numArgs == 1 ? SymFindFolderName() : folder;
		
			// display actual folder on status line
			sprintf (out, "searching %s", varFolder);
			ST_helpMsg (out);
		
			// variable is not a folder
      if (IsFile (symptr) && 
      IsFuncOrProg (symptr)) {
      	// variable contains search string
      	if (MultiExprContStr (var, str)) {	
        	// add matiching variable to list
					sprintf (out, "%s\\%s", varFolder, symptr->name);
						
        	push_ANSI_string (out);   	
      	}
      	
      	numVars++;
      }
        
      // get next variable from folder
    } while ((symptr = SymFindNext ()));
  }
  	
  // unlock home folder
  FolderOp (SYMSTR (HOME), FOP_UNLOCK);
  	
  // end of list
  push_LIST_TAG ();
  	
  // display number of variables searched on status line
	sprintf (out, "%d variables searched", numVars);
	ST_helpMsg (out);
}
