/*
		UnresRef - find functions/programs containing missing references
		Example: UnresRef("folder")
		if the parameter is ommited all folders are searched
		
    Copyright (C) 2013  Ralph Schenn

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <alloc.h>
#include <args.h>
#include <error.h>
#include <estack.h>
#include <statline.h>
#include <string.h>
#include <vat.h>
#include "Memory.h"
#include "MultiExpr.h"
#include "Stack.h"

#define SMALL_BUF 30

// this should be enough for at least 10 references
#define BIG_BUF 100

void UnresRefMat(const char *pathCaller, HANDLE refs[], unsigned short length)
/* print matrix with unresolved references 
[["folder_caller\var_caller","folder_called\var_called",...]...]*/
{
	char buf1[BIG_BUF] = ""; // ??? save space
	char buf2[BIG_BUF] = "";
	
	// reference counter
	unsigned short i;
  
  // loop over references in caller
  for (i = 0; i < length; i++) {
  	// get pointer to next reference path
  	const char *nextRef = HeapDeref (refs[i]);
  	
  	// called function/program is missing
  	if (!VarExists (nextRef)) {  		
			// found first missing reference
			if (strlen (buf1) == 0) {
  			strcpy (buf1, nextRef);
  		// enough space for next missing reference
  		} else if (strlen (buf1) + strlen (nextRef) < BIG_BUF - 5) {
  			// append missing reference
  			sprintf (buf2, "%s,%s", buf1, nextRef);
  			strcpy (buf1, buf2);
  		// no more space for additional missing references
  		} else {
  			// mark overflow
  			sprintf (buf2, "%s,...", buf1);
  			strcpy (buf1, buf2);  				
  		}
  	}
  }
  
  // caller contains at least one missing reference
  if (strlen (buf1) > 0) {
	  // end of row
	  push_END_TAG ();
  	
  	// push missing references in caller
  	push_ANSI_string (buf1);

  	// push the caller
  	push_ANSI_string (pathCaller);  	
  	
	  // beginning of row
	  push_LIST_TAG ();	
  }
}

void _main(void)
{	
	#define HOME "\x7F"
	
	// flags for variable search
	unsigned short searchFlags;

	// parent folder of variable
	const char *varFolder;

	// search folder name
	SYM_STR symName;	

	// pointer to variable
	SYM_ENTRY *symptr;

	// number of variables which have been searched
	unsigned short numVars = 0;

	// output buffer
	char out[SMALL_BUF];
	
	CESI argptr = top_estack;
	
	// array with references in caller
	HANDLE refs = H_NULL;

	// check number of arguments
	char numArgs = CheckNumArgsInt (0, 1);

	TRY		
		// search through all folders
		if (numArgs == 0) {
			// no folder name
			symName = NULL;
			
			// search all folders
			searchFlags = FO_RECURSE;
		// search only in  folder given in second argument
		} else {
			// get pointer to folder name from estack
			// (this must be tagged format, not C format)
			symName = GetSymstrArg (argptr);
			
			// search only in specified folder
			searchFlags = FO_SINGLE_FOLDER;
		}
	ONERR
		// invalid folder name
		ER_throw (ER_ARG_MUST_BE_STRING);
	ENDTRY

	TRY
		// lock home folder
		FolderOp (SYMSTR (HOME), FOP_LOCK);
		
		// find first variable
		symptr = SymFindFirst (symName, searchFlags);
		
		// remove arguments from estack
		// (symptr points to the estack, so this should be done 
		// only after using symptr)
		CleanStack ();
		
		// at least one variable exists
	  if (symptr != NULL) {
			// end of matrix
	  	push_END_TAG ();

    	// loop over variables in folder(s)
	    do {
				// determine actual folder
				varFolder = SymFindFolderName();
			
				// display actual folder on status line
				sprintf (out, "searching %s", varFolder);
				ST_helpMsg (out);
								
				// variable is not a folder
	      if (IsFile (symptr)) {
					// get references in caller
					unsigned short numRefs = GetReferences (&refs, symptr);
					
					// caller contains at least one reference
					if (numRefs > 0) {
	        	// path of caller
						sprintf (out, "%s\\%s", varFolder, symptr->name);
						
						HANDLE *refPtr = HeapDeref (refs);
						
						// check if there are missing references
						UnresRefMat (out, refPtr, numRefs);	
						
						// free handles to strings with reference paths
	  				StringHandleArrayFree(refPtr, numRefs);
					}
					
					numVars++;
	      }
	        	        
	      // get next variable from folder(s)
	      symptr = SymFindNext ();
	    } while (symptr != NULL);
	    
	  	// begin of matrix
	  	push_LIST_TAG ();
	  // folder not found
	  } else {
	  	ER_throw (ER_INVALID_FOLDER_NAME);
	  }
	  
	  // display number of variables searched on status line
		sprintf (out, "%d variables searched", --numVars);
		ST_helpMsg (out);
  FINALLY
	  // unlock home folder
	  FolderOp (SYMSTR (HOME), FOP_UNLOCK);
	  	  
	  // at least one variable was found
	  if (refs != H_NULL) {
	  	// free array with references in caller
	  	HeapFree (refs);
	  }
	ENDFINAL
}
